/*
INIT
Copyright (c) 2018, Greg Sydney-Smith
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

2018-09-02. gss. Initial version. Source to sydneysmith.com.
*/
/* BDSC 1.60 has unix-like stdio. Earlier vers are incompatible. */
#include <stdio.h>

#define MAXSIZ          8000
#define malloc(x)       alloc(x)
#define FMT             struct _fmt


struct _fmt {
    int map1[30];
    int map2[30];
    int map3[30];
    int map4[30];
    int tkIdx;          /* Track (index into map2) */
    int sdIdx;          /* Side ... */
    int scIdx;          /* Sector ... */
    int daOfs;          /* Data (bytes into expanded map2) */
} ;


FMT sd8, dd8, sd5, dd5, sd48, sd45;
int  lastTrack;
int  lastSide;
char asmWrTk[12+1];
char img[MAXSIZ];


char *strncpy(s2,s1,n)
char *s2;
char *s1;
int n;
{
    int i;
    for (i=0; i<n && s1[i]; i++) s2[i]=s1[i];
    return s2;
}


        /* **** 16FDC Interface **** */


int drvSelect(drv,sz)
int drv;
int sz;
{
    int v;

    /* Select Drive (no autowait) */
    v= 0x20;                    /* motor on */
    if (sz ==8) v |= 0x10;      /* 8" maxi */
    if (drv==0) v |= 1;
    if (drv==1) v |= 2;
    if (drv==2) v |= 4;
    if (drv==3) v |= 8;
    outp(0x34,v);               /* drv select */

    /* Select Side 0 */
    outp(0x04,0xff);            /* side 0 */

    /* Select Track 0 */
    outp(0x30,0x08);            /* restore(0) + headload(8) */
    while ((inp(0x34) & 1)==0) ; /* wait for EOJ bit */
    if (inp(0x30) & 0x98) { printf("\nERROR restore to track 0 failed\n"); return 1; }

    lastTrack=0;
    lastSide =0;
    return 0;
}

/* INIT 00.04 uses 0x58
   INIT 02.71 uses 0x5B */
int drvStep() {
    outp(0x30,0x58);            /* stepIn + updateTrack + headload */
    while ((inp(0x34) & 1)==0) ; /* wait for EOJ bit */
    if (inp(0x30) & 0x98) { printf("\nERROR step in failed\n"); return 1; }
    return 0;
}

int drvWriteTrack(drv,sz,dd,track,side,buf)
int drv;
int sz;
int dd;
int track;
int side;
char *buf;
{
    int v;

    /* Select Drive (mainly just turn autowait on) */
    v  = 0x20;                  /* motor on */
    v |= 0x80;                  /* autowait */
    if (dd   ) v |= 0x40;       /* double density */
    if (sz==8) v |= 0x10;       /* 8" maxi */
    if (drv==0) v |= 1;
    if (drv==1) v |= 2;
    if (drv==2) v |= 4;
    if (drv==3) v |= 8;
    outp(0x34,v);               /* drv select */

    /* Select Side */
    v= (side==0)? 0xff : 0xfd;
    outp(0x04,v);

    /* use asm for most of this. needs speed
       0. HL=buf, C=0x33
       1. outp(0x30,0xF4); // write track command (have autowait on before this)
       2. check for EOJ ie inp(0x30)&1, if so done else
       3. OUTI (to send (HL)->port 0x33 then INC HL)
       4. go step 2
       asm {
	; LD	HL,(IMGADDR)	; L0103
	; LD	C,33H		; data port
	; LD	A,0F4H		; write track cmd
	OUT	(30H),A		; cmd port
iL09DC:	IN	A,(34H)
	RRA
	JR	C,iL09E5
	OUTI
	JR	L09DC
iL09E5	RET
       
       }
       asmWrTk=initb("211,48,219,52,31,56,4,237,163,24,247,201");
    */
    call(asmWrTk,0xf4,buf,0x0033,0);    /* (addr,a,hl,bc,de) */
    v= inp(0x30);
    if (v & 0x40) { printf("\nERROR disk is write protected\n"); return 1; }
    if (v & 0xe4) { printf("\nERROR write track failed\n"); return 1; }
    return 0;
}


        /* **** Generic from here down **** */


int writeTrack(drv, sz, dd, track, side, buf)
int drv;
int sz;
int dd;
int track;
int side;
char *buf;
{
    printf("\rWriting track %2d side %d",track,side);

    if (track!=lastTrack) {
        if (drvStep() != 0) return 1;
    }

    if (drvWriteTrack(drv,sz,dd,track,side,buf) != 0) return 1;
    lastTrack= track;
    lastSide = side;
    return 0;
}


int makeChunk(map, buf)
int map[];
char *buf;
{
    int siz, i, j, len, c;

    for (siz=0, i=0; 1; i++) {
        len= map[2*i+0];
        c  = map[2*i+1];
        if (len < 0) break;
        for (j=0; j<len; j++) *buf++ = c;
        siz+=len;
    }
    return siz;
}

int makeTrack(fmt, id, track, side, spt, buf)
FMT *fmt;
char *id;
int track;
int side;
int spt;
char *buf;
{
    int len, sector;

    buf += makeChunk(fmt->map1,buf);            /* make track start */

    fmt->map2[fmt->tkIdx]= track;               /* make sector 1 */
    fmt->map2[fmt->sdIdx]= side;
    fmt->map2[fmt->scIdx]= 1;
    len= makeChunk(fmt->map2,buf);
    if (track==0 && side==0)
        strncpy(buf+fmt->daOfs+120,id,strlen(id));
    buf += len;

    for (sector=2; sector<=spt; sector++) {
        fmt->map2[fmt->scIdx]= sector;          /* set next sector number */
        buf += makeChunk(fmt->map3,buf);        /* between sectors */
        buf += makeChunk(fmt->map2,buf);        /* make sector */
    }

    buf += makeChunk(fmt->map4,buf);        /* make track end */
    return 0;
}


int initDsk(drv, id, sz, tracks, sides, dd0, dd1, spt0, spt1, fmt0, fmt1)
int drv;
char *id;
int sz;
int tracks;
int sides;
int dd0;
int dd1;
int spt0;
int spt1;
FMT *fmt0;
FMT *fmt1;
{
    int track, side;

    drvSelect(drv,sz);

    /* Tk 0 Side 0 */
    makeTrack(fmt0,id,0,0,spt0,img);
    writeTrack(drv,sz,dd0,0,0,img);
/* return 0; */

    /* Remaining tracks */
    for (track=0; track<tracks; track++) {
        for (side=0; side<sides; side++) {
            if (track==0 && side==0) continue; /* already done */
            makeTrack(fmt1,"",track,side,spt1,img);
            if (writeTrack(drv,sz,dd1,track,side,img) != 0) return 1;
        }
    }
    return 0;
}


int    ans_x;
int    ans_n;
char **ans_v;


int askChr(msg)
char *msg;
{
    int c;
    char line[132+1];

    if (ans_x < ans_n) strcpy(line,ans_v[ans_x++]);
    else {
        printf("%s? ",msg);
        fgets(line,132,stdin);
    }

    c=line[0];
    if ('a'<=c && c<='z') c= c-'a'+'A';
    return c;
}

int askNum(msg,dflt)
char *msg;
char *dflt;
{
    int n;
    char line[132+1];

    if (ans_x < ans_n)
        if ((n=atoi(ans_v[ans_x++])) > 0)
            return n;
    ans_n=0;

    printf("%s [%s]? ",msg,dflt);
    fgets(line,132,stdin);
    for (n=strlen(line)-1; n>=0 && line[n]<=' '; n--) line[n]='\0';
    if (line[0]=='\0') return atoi(dflt);
    return atoi(line);
}

int initmain()
{
    FMT *fmt0, *fmt1;
    int  drv, sz, tracks, sides, dd0, dd1, spt0, spt1, sd0, use4, bt;
    char dtrk[3+1], dsec[3+1], id[6+1];

    printf("INIT Ver 00.01\n");
    printf("(c) 2018 sydneysmith.com\n\n");

    drv  = askChr("Disk to initialize (A-D) [quit]","ABCD");
    if (drv<'A' || drv>'D') return 0;
    drv-='A'; /* A-D -> 0-3 */

    sz   = askChr("Disk size                (5 or 8) [8]") == '5'? 5 : 8;
    sides= askChr("Single or double sided   (S or D) [S]") == 'D'? 2 : 1;
    dd1  = askChr("Single or double density (S or D) [S]") == 'D';

    dd0  = 0;
    use4 = 0;
    bt   = 1;
    if (sz==8) { tracks=77; spt0=26; spt1= dd1? 16 : 26; strcpy(id,"LG"); }
    if (sz==5) { tracks=40; spt0=18; spt1= dd1? 10 : 18; strcpy(id,"SM"); }
    strcat(id,sides==1?"SS":"DS");
    strcat(id,dd1     ?"DD":"SD");
    sprintf(dtrk,"%d",tracks);
    sprintf(dsec,"%d",spt1);

/*
    if (askChr("\nCreate a custom format   (Y or N) [N]")=='Y') {
        printf("If your OS and hardware supports it, you may change the following.\n");
        if ( dd1) dd0 = askChr("Single density track 0   (Y or N) [Y]")!='N';
        tracks        = askNum("Tracks                   (35-128)",dtrk);
        spt1          = askNum("Sectors                   (8-128)",dsec);
        if (tracks<35) tracks=35; if (tracks>128) tracks=128;
        if (spt1  <8 ) spt1  = 8; if (spt1  >128) spt1  =128;
        if (!dd1) use4= askChr("Create 4FDC format disk  (Y or N) [N]")=='Y';
        if (tracks!=atoi(dtrk) || spt1!=atoi(dsec) || use4) bt=0;
        if ( bt ) bt  = askChr("Write type to disk       (Y or N) [Y]")!='N';
        if (!bt ) id[0]='\0';
    }
*/

    spt0= spt1;
    dd0 = dd1;
    if (sz==8) {
        fmt1= (dd1)? dd8 : (use4)? sd48 : sd8;
        fmt0= fmt1;
        if (!dd0) { spt0=26; fmt0= (use4)? sd48 : sd8; }
    }
    if (sz==5) {
        fmt1= (dd1)? dd5 : (use4)? sd45 : sd5;
        fmt0= fmt1;
        if (!dd0) { spt0=18; fmt0= (use4)? sd45 : sd5; }
    }

    printf("\nAll existing data on the disk will be lost.\n");
    if (askChr("Proceed (Y/N) [N]")=='Y') {
        printf("\nFormatting ...\n");
        initDsk(drv,id,sz,tracks,sides,dd0,dd1,spt0,spt1,fmt0,fmt1);
        printf("\nDone.");
    }
    return 0;
}

int main(argc,argv)
int argc;
char *argv[];
{

    /* preload any provided answers */
    ans_x= 0;
    ans_n= argc-1;
    ans_v= argv+1;

    /* BDS C does not support initializers so we have to:
       0x4E=78 0xE5=229 0xF7=247 0xFB=251 0xFC=252 0xFD=253 0xFE=254 */

    /* 8" SD */
    initw(sd8.map1,"40,255,6,0,1,252,26,255,6,0,-1");
    initw(sd8.map2,"1,254,1,0,1,0,1,0,1,0,1,247,11,255,6,0,1,251,128,229,1,247,-1");
    initw(sd8.map3,"27,255,6,0,-1");
    initw(sd8.map4,"1051,255,-1");
    sd8.tkIdx=3; sd8.sdIdx=5; sd8.scIdx=7; sd8.daOfs=24;

    /* 5" SD */
    initw(sd5.map1,"10,255,4,0,-1");
    initw(sd5.map2,"1,254,1,0,1,0,1,0,1,0,1,247,11,255,6,0,1,251,128,229,1,247,-1");
    initw(sd5.map3,"11,255,4,0,-1");
    initw(sd5.map4,"1051,255,-1");
    sd5.tkIdx=3; sd5.sdIdx=5; sd5.scIdx=7; sd5.daOfs=38;

    /* 8" DD */
    initw(dd8.map1,"80,78,12,0,3,246,1,252,50,78,12,0,3,245,-1");
    initw(dd8.map2,"1,254,1,0,1,0,1,0,1,2,1,247,22,78,12,0,3,245,1,251,512,229,1,247,-1");
    initw(dd8.map3,"54,78,12,0,3,245,-1");
    initw(dd8.map4,"1077,78,-1");
    dd8.tkIdx=3; dd8.sdIdx=5; dd8.scIdx=7; dd8.daOfs=44;

    /* 5" DD */
    initw(dd5.map1,"24,78,8,0,3,245,-1");
    initw(dd5.map2,"1,254,1,0,1,0,1,0,1,2,1,247,22,78,12,0,3,245,1,251,512,229,1,247,-1");
    initw(dd5.map3,"1,255,45,254,8,0,3,245,-1");
    initw(dd5.map4,"1,255,1024,78,-1");
    dd5.tkIdx=3; dd5.sdIdx=5; dd5.scIdx=7; dd5.daOfs=79;

    /* 8" 4FDC SD */
    initw(sd48.map1,"46,0,1,252,32,0,-1");
    initw(sd48.map2,"1,254,1,0,1,0,1,0,1,0,1,247,17,0,1,251,128,229,1,247,-1");
    initw(sd48.map3,"33,0,-1");
    initw(sd48.map4,"1060,0,-1");
    sd48.tkIdx=3; sd48.sdIdx=5; sd48.scIdx=7; sd48.daOfs=24;

    /* 5" 4FDC SD */
    initw(sd45.map1,"7,255,4,0,-1");
    initw(sd45.map2,"1,254,1,0,1,0,1,0,1,0,1,247,11,255,6,0,1,251,128,229,1,247,-1");
    initw(sd45.map3,"8,255,4,0,-1");
    initw(sd45.map4,"1060,0,-1");
    sd45.tkIdx=3; sd45.sdIdx=5; sd45.scIdx=7; sd45.daOfs=24;

    /* asm for z80 Write Track (tight loop) */
    initb(asmWrTk,"211,48,219,52,31,56,4,237,163,24,247,201");

    initmain();
    return 0;
}

;
