/*
INIT2
Copyright (c) 2018, Greg Sydney-Smith
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

2018-09-23. gss. Initial version. Source to sydneysmith.com.
*/
/* BDSC 1.60 has unix-like stdio. Earlier vers are incompatible. */
#include <stdio.h>

#define MAXSIZ          10000
#define malloc(x)       alloc(x)
#define FMT             struct _fmt


struct _fmt {
    int map1[30];
    int map2[30];
    int map3[30];
    int map4[30];
    int tkIdx;          /* Track (index into map2) */
    int sdIdx;          /* Side ... */
    int scIdx;          /* Sector ... */
    int daOfs;          /* Data (bytes into expanded map2) */
} ;


FMT sd8, dd8, sd5, dd5;
int  lastTrack;
int  lastSide;
char img[MAXSIZ];


char *strncpy(s2,s1,n)
char *s2;
char *s1;
int n;
{
    int i;
    for (i=0; i<n && s1[i]; i++) s2[i]=s1[i];
    return s2;
}


/*

**** GEN2 Interface ****
FDCDISK		EQU	30H
FDCTRKH		EQU	31H
FDCTRKL		EQU	32H
FDCSECT		EQU	33H
FDCDMAH		EQU	35H
FDCDMAL		EQU	36H
FDCCMD		EQU	37H
*/

int drvSelect(drv)
int drv;
{
    /* Select Drive */
    outp(0x30,drv);

    /* Select Track 0 */
    outp(0x31,0x00);
    outp(0x32,0x00);

    lastTrack=0;
    lastSide =0;
    return 0;
}

int drvStep(ds)
int ds;
{
    /* nothing required - done in drvWriteTrack() */
    return 0;
}

int drvWriteTrack(drv,ds,track,side,buf)
int drv;
int ds;
int track;
int side;
char *buf;
{
    int v, n;


    n= (ds)? track*2+side : track;
    outp(0x31,n/256);
    outp(0x32,n%256);

    n= buf;
    outp(0x35,n/256);
    outp(0x36,n%256);
    outp(0x37,2);       /* write track */
    v=inp(0x37);
    if (v) { printf("\nERROR write track failed\n"); return 1; }
    return 0;
}


        /* **** Generic from here down **** */


int writeTrack(drv, ds, track, side, buf)
int drv;
int ds;
int track;
int side;
char *buf;
{
    printf("\rWriting track %2d side %d",track,side);

    if (track!=lastTrack) {
        if (drvStep(ds) != 0) return 1;
    }

    if (drvWriteTrack(drv,ds,track,side,buf) != 0) return 1;
    lastTrack= track;
    lastSide = side;
    return 0;
}


int makeChunk(map, buf)
int map[];
char *buf;
{
    int siz, i, j, len, c;

    for (siz=0, i=0; 1; i++) {
        len= map[2*i+0];
        c  = map[2*i+1];
        if (len < 0) break;
        for (j=0; j<len; j++) *buf++ = c;
        siz+=len;
    }
    return siz;
}

int makeTrack(fmt, id, track, side, spt, buf)
FMT *fmt;
char *id;
int track;
int side;
int spt;
char *buf;
{
    int len, sector;

    buf += makeChunk(fmt->map1,buf);            /* make track start */

    fmt->map2[fmt->tkIdx]= track;               /* make sector 1 */
    fmt->map2[fmt->sdIdx]= side;
    fmt->map2[fmt->scIdx]= 1;
    len= makeChunk(fmt->map2,buf);
    if (track==0 && side==0)
        strncpy(buf+fmt->daOfs+120,id,strlen(id));
    buf += len;

    for (sector=2; sector<=spt; sector++) {
        fmt->map2[fmt->scIdx]= sector;          /* set next sector number */
        buf += makeChunk(fmt->map3,buf);        /* between sectors */
        buf += makeChunk(fmt->map2,buf);        /* make sector */
    }

    buf += makeChunk(fmt->map4,buf);        /* make track end */
    return 0;
}


int initTrk0(drv,spt,fmt,id)
int drv;
int spt;
FMT *fmt;
char *id;
{
    drvSelect(drv);
    makeTrack(fmt,id,0,0,spt,img);
    writeTrack(drv,0,0,0,img);     /* ds doesn't matter on tk0 hd0 */
    return 0;
}

int initTrks(drv,tracks,sides,spt,fmt)
int drv;
int tracks;
int sides;
int spt;
FMT *fmt;
{
    int ds,track, side;

    /* Remaining tracks */
    ds= (sides==2);
    for (track=0; track<tracks; track++) {
        for (side=0; side<sides; side++) {
            if (track==0 && side==0) continue; /* already done */
            makeTrack(fmt,"",track,side,spt,img);
            if (writeTrack(drv,ds,track,side,img) != 0) return 1;
        }
    }
    return 0;
}


int    ans_x;
int    ans_n;
char **ans_v;


int askChr(msg)
char *msg;
{
    int c;
    char line[132+1];

    if (ans_x < ans_n) strcpy(line,ans_v[ans_x++]);
    else {
        printf("%s? ",msg);
        fgets(line,132,stdin);
    }

    c=line[0];
    if ('a'<=c && c<='z') c= c-'a'+'A';
    return c;
}

int askDrv() {
    int v;
    if ((v=askChr("Disk to initialize       (A-D) [quit]")) < ' ') v= -1;
    return (v<'A' || v>'D')? -1 : v-'A';
}

int askSize() {
    int v;
    if ((v=askChr("Disk size              (5,8 or H) [8]")) < ' ') v='8';
    return v;
}

int askSides() {
    int v;
    if ((v=askChr("Single or double sided   (S or D) [D]")) < ' ') v='D';
    return v;
}

int askDensity() {
    int v;
    if ((v=askChr("Single or double density (S or D) [D]")) < ' ') v='D';
    return v;
}

int askTrack0() {
    int v;
    if ((v=askChr("Single density track 0   (Y or N) [N]")) < ' ') v='N';
    return v;
}

int askProceed() {
    int v;
    printf("\nAll existing data on the disk will be lost.\n");
    if ((v= askChr("Proceed (Y/N) [N]")) < ' ') v='N';
    return v;
}


int initmain()
{
    int  state, drv, sz, ds, dd, d0, ok;

    printf("INIT Ver 00.01 for G2FDC\n");
    printf("(c) 2018 Greg Sydney-Smith\n\n");

    for (dd='S', state=1; state; ) {
        drv=askDrv(); if (drv<0) break;
        sz =askSize();
        if (sz!='H') ds= askSides();
        if (sz!='H') dd= askDensity();
        if (dd=='D') d0= askTrack0();
        ok =askProceed(); if (ok!='Y') break;
        printf("\n");

        if (sz=='8' && ds=='S' && dd=='S')
            { initTrk0(drv,26,sd8,"LGSSSD_A"); initTrks(drv,77,1,  26,sd8); }
        if (sz=='8' && ds=='S' && dd=='D' && d0=='Y')
            { initTrk0(drv,26,sd8,"LGSSDD_E"); initTrks(drv,77,1,  16,dd8); }
        if (sz=='8' && ds=='S' && dd=='D' && d0=='N')
            { initTrk0(drv,16,dd8,"8_SSDD_C"); initTrks(drv,77,1,  16,dd8); }
        if (sz=='8' && ds=='D' && dd=='S')
            { initTrk0(drv,26,sd8,"LGDSSD_B"); initTrks(drv,77,2,  26,sd8); }
        if (sz=='8' && ds=='D' && dd=='D' && d0=='Y')
            { initTrk0(drv,26,sd8,"LGDSDD_F"); initTrks(drv,77,2,  16,dd8); }
        if (sz=='8' && ds=='D' && dd=='D' && d0=='N')
            { initTrk0(drv,16,dd8,"8_DSDD_D"); initTrks(drv,77,2,  16,dd8); }

        if (sz=='5' && ds=='S' && dd=='S')
            { initTrk0(drv,18,sd5,"SMSSSD_a"); initTrks(drv,40,1,  18,sd5); }
        if (sz=='5' && ds=='S' && dd=='D' && d0=='Y')
            { initTrk0(drv,18,sd5,"SMSSDD_e"); initTrks(drv,40,1,  10,dd5); }
        if (sz=='5' && ds=='S' && dd=='D' && d0=='N')
            { initTrk0(drv,10,dd5,"5_SSDD_c"); initTrks(drv,40,1,  10,dd5); }
        if (sz=='5' && ds=='D' && dd=='S')
            { initTrk0(drv,18,sd5,"SMDSSD_b"); initTrks(drv,40,2,  18,sd5); }
        if (sz=='5' && ds=='D' && dd=='D' && d0=='Y')
            { initTrk0(drv,18,sd5,"SMDSDD_f"); initTrks(drv,40,2,  10,dd5); }
        if (sz=='5' && ds=='D' && dd=='D' && d0=='N')
            { initTrk0(drv,10,dd5,"5_DSDD_d"); initTrks(drv,40,2,  10,dd5); }

        if (sz=='H')
            { initTrk0(drv,16,dd8,"HDSSDD_H"); initTrks(drv,1024,1,16,dd8); }

        printf("\nDone.\n\n");
    }
    

    return 0;
}

int main(argc,argv)
int argc;
char *argv[];
{

    /* preload any provided answers */
    ans_x= 0;
    ans_n= argc-1;
    ans_v= argv+1;

    /* BDS C does not support initializers so we have to:
       0x4E=78 0xE5=229 0xF7=247 0xFB=251 0xFC=252 0xFD=253 0xFE=254 */

    /* 8" SD */
    initw(sd8.map1,"40,255,6,0,1,252,26,255,6,0,-1");
    initw(sd8.map2,"1,254,1,0,1,0,1,0,1,0,1,247,11,255,6,0,1,251,128,229,1,247,-1");
    initw(sd8.map3,"27,255,6,0,-1");
    initw(sd8.map4,"1051,255,-1");
    sd8.tkIdx=3; sd8.sdIdx=5; sd8.scIdx=7; sd8.daOfs=24;

    /* 5" SD */
    initw(sd5.map1,"10,255,4,0,-1");
    initw(sd5.map2,"1,254,1,0,1,0,1,0,1,0,1,247,11,255,6,0,1,251,128,229,1,247,-1");
    initw(sd5.map3,"11,255,4,0,-1");
    initw(sd5.map4,"1051,255,-1");
    sd5.tkIdx=3; sd5.sdIdx=5; sd5.scIdx=7; sd5.daOfs=38;

    /* 8" DD */
    initw(dd8.map1,"80,78,12,0,3,246,1,252,50,78,12,0,3,245,-1");
    initw(dd8.map2,"1,254,1,0,1,0,1,0,1,2,1,247,22,78,12,0,3,245,1,251,512,229,1,247,-1");
    initw(dd8.map3,"54,78,12,0,3,245,-1");
    initw(dd8.map4,"1077,78,-1");
    dd8.tkIdx=3; dd8.sdIdx=5; dd8.scIdx=7; dd8.daOfs=44;

    /* 5" DD */
    initw(dd5.map1,"24,78,8,0,3,245,-1");
    initw(dd5.map2,"1,254,1,0,1,0,1,0,1,2,1,247,22,78,12,0,3,245,1,251,512,229,1,247,-1");
    initw(dd5.map3,"1,255,45,254,8,0,3,245,-1");
    initw(dd5.map4,"1,255,1024,78,-1");
    dd5.tkIdx=3; dd5.sdIdx=5; dd5.scIdx=7; dd5.daOfs=79;

    initmain();
    return 0;
}


